/* $Id: automat.c,v 1.32 2003/01/04 11:26:26 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include "util.h"
#include "automat.h"
#include "lists.h"
#include "dstring.h"

/// \section[WHAT ARE THOSE PDTS?]
//
// The following is just a brief introduction. If you're completely unfamiliar with
// the automata theory, you'd better learn something more from other sources.
//
// Push-down transducers are a class of \em2[formal automata][,] abstract mathematical
// structures which may change their internal state and produce some output depending on 
// the input. One of the most generic (and most well-known) examples of automata is 
// the \em2[Turing's machine][.]
//
// Strictly speaking, the term automata is properly applicable only to those structures
// which has no output and can only decide whether a given input sequence belong to a
// certain class or not (the class which is usually called a \em[language] 
// is said to be \em[recognized] by the automaton. Those devices which can produce some
// output are properly called \em[transducers] but the latter term is not very wide spread,
// so further we will use the both as mutual synomyms.
//
// The most primitive automata class is \em2[finite-state automata][.] Their state and output
// are determined by a single character of the input and the current state, the sets of input 
// characters and inner states both being \em2[finite][.]
//
// \em[Push-down automata] have in addition an internal stack of unlimited length. The topmost
// character in the stack determines the behaviour of an automaton alongside with its previous 
// state and the input character. After changing the state and emitting the output, the topmost
// stack character is replaced by an arbitrary (poss. empty) string of characters.
// The set of tasks that can be modelled with push-down automata/transducers is much
// larger than that of finite-state automata so the latter should be clearly distinguished from
// the former. However, in practice, push-down automata are often called finite-state.
//
// Automata are an ideal device to deal with complex string transformations. Well-known 
// regular expressions are nothing but another notation for finite-state (outputless) automata.
// However, they may be quite hard to learn, so they are not very popular nowadays 
// (I mean, as a tool for text processing). I personally know the only project which uses
// this conception -- namely, Omega project by John Plaice anf Yannis Haralambous.

node *use_node(node *n)
{
	n->ref_cnt++;
	return n;
}

static node *create_variable_node(int var)
{
	node *n = ALLOC(node);
	n->type = node_variable;
	n->x.val = var;
	return n;
}

static node *create_value_node(int val)
{
	node *n = ALLOC(node);
	n->type = node_const;
	n->x.val = val;
	return n;
}

static node *create_selection_node(range *r, node *val)
{
	node *n = ALLOC(node);
	n->type = node_selection;
	n->x.sel.r = r;
	n->x.sel.right = use_node(val);
	return n;
}

static node *create_rangecheck_node(range *r, node *val)
{
	node *n = ALLOC(node);
	n->type = node_rangecheck;
	n->x.sel.r = r;
	n->x.sel.right = use_node(val);
	return n;
}

/*
static node *create_autoseq_node(autoseq *seq, node *val)
{
	node *n = ALLOC(node);
	n->type = node_autoseq;
	n->x.aseq.seq = seq;
	n->x.aseq.right = use_node(val);
	return n;
}
*/

static node *create_assign_node(int reg, node *left)
{
	node *n = ALLOC(node);
	n->type = node_assign;
	n->x.a.regno = reg;
	n->x.op.left = use_node(left);
	return n;
}


static node *create_binop_node(binop_t opf, node *left, node *right)
{
	node *n = ALLOC(node);
	n->type = node_binop;
	n->x.op.opf = opf;
	n->x.op.left = use_node(left);
	n->x.op.right = use_node(right);
	return n;
}

static node *create_unop_node(binop_t opf, node *left)
{
	node *n = ALLOC(node);
	n->type = node_unop;
	n->x.op.opf = opf;
	n->x.op.left = use_node(left);
	n->x.op.right = NULL;
	return n;
}

/// \section[EXPRESSIONS]
// Automata expressions are much like C expressions. However, in the current version
// they have less precedence levels that the latter.
// Expressions may contain the following operators (in descending precedence)
// \item \em[- ! ~~] -- unary minus, logical NOT, bitwise inversion
// \item \em[* / %] -- multiplication, division, remainder
// \item \em[+ -] -- addition, subtraction
// \item \em[& | ^ << >>] -- bitwise AND, OR, XOR, left/right shift
// \item \em[= !\|= < > >\|= <\|=] -- equal, not equal, less, greater, greater-or-equal,
// less-or-equal (mind a single \em[=] not like in C)
// \item \em[&& ||] -- logical AND, OR
// \item \em[,] -- sequential execution
// \endlist
//
// There are yet three special operators (`@', `:', and application) which will
// be described later.
// Parentheses may be used in a common way.
// Primitive expressions are integral constant (\ref[strtol(3)] syntax), 
// character constants (C syntax),
// \em[register references], and macro names


int binop_add(int arg1, int arg2) { return arg1 + arg2; }
int binop_sub(int arg1, int arg2) { return arg1 - arg2; }
int binop_mul(int arg1, int arg2) { return arg1 * arg2; }
int binop_div(int arg1, int arg2) { if(arg2) return arg1 / arg2; else recover_error("binop_div: zero division"); return 0; }
int binop_rem(int arg1, int arg2) { if(arg2) return arg1 % arg2; else recover_error("binop_rem: zero division"); return 0; }
int binop_shiftl(int arg1, int arg2) { return arg1 << arg2; }
int binop_shiftr(int arg1, int arg2) { return arg1 >> arg2; }
int binop_and(int arg1, int arg2) { return arg1 & arg2; }
int binop_or(int arg1, int arg2) { return arg1 | arg2; }
int binop_xor(int arg1, int arg2) { return arg1 ^ arg2; }
int binop_land(int arg1, int arg2) { return arg1 && arg2; }
int binop_lor(int arg1, int arg2) { return arg1 || arg2; }
int binop_eq(int arg1, int arg2) { return arg1 == arg2; }
int binop_neq(int arg1, int arg2) { return arg1 != arg2; }
int binop_less(int arg1, int arg2) { return arg1 < arg2; }
int binop_greater(int arg1, int arg2) { return arg1 > arg2; }
int binop_le(int arg1, int arg2) { return arg1 <= arg2; }
int binop_ge(int arg1, int arg2) { return arg1 >= arg2; }
int binop_seq(int arg1, int arg2) { return arg2; }

int unop_inv(int arg1, int arg2) { return ~arg1; }
int unop_lnot(int arg1, int arg2) { return !arg1; }
int unop_neg(int arg1, int arg2) { return -arg1; }

int select_value(range *r, int value)
{
	for(; r->repeat > 0; r++)
	{
		if(value >= r->base && value < r[1].base)
		{
			if(r->dither > 0)
				return (value - r->base) * r->dither % (r->end - r->start + 1) + r->start;
			else return r->start - (value - r->base) * -r->dither % (r->start - r->end + 1);
		}
	}
	recover_error("value out of range: %d", value);
	return 0;
}

int in_range(range *r, int value)
{
	for(; r->repeat > 0; r++)
	{
		if((value - r->start) / r->dither >= 0 && (value - r->end) / r->dither <= 0 && !((value - r->start) % r->dither))
			return r->base + (value - r->start) / r->dither;
	}
	return -1;
}

int eval_node(node *n, reg_block_t arg)
{
	switch(n->type)
	{
		case node_const:
			return n->x.val;
		case node_variable:
			return arg[n->x.val];
		case node_assign:
			return arg[n->x.a.regno] = eval_node(n->x.a.left, arg);
		case node_binop:
			return n->x.op.opf(eval_node(n->x.op.left, arg), eval_node(n->x.op.right, arg));
		case node_unop:
			return n->x.op.opf(eval_node(n->x.op.left, arg), 0);
		case node_selection:
			return select_value(n->x.sel.r, eval_node(n->x.sel.right, arg));
		case node_rangecheck:
			return in_range(n->x.sel.r, eval_node(n->x.sel.right, arg));
		default:
			fatal_error("eval_node: bad node type (%d)", n->type);
	}
	return 0;
}

node *copy_node(node *n)
{
	switch(n->type)
	{
		case node_const:
			return create_value_node(n->x.val);
		case node_variable:
			return create_variable_node(n->x.val);
		case node_assign:
			return create_assign_node(n->x.a.regno, copy_node(n->x.a.left));
		case node_binop:
			return create_binop_node(n->x.op.opf, copy_node(n->x.op.left), copy_node(n->x.op.right));
		case node_unop:
			return create_unop_node(n->x.op.opf, copy_node(n->x.op.left));
		case node_selection:
			return create_selection_node(n->x.sel.r, copy_node(n->x.sel.right));
		case node_rangecheck:
			return create_rangecheck_node(n->x.sel.r, copy_node(n->x.sel.right));
		default:
			fatal_error("eval_node: bad node type (%d)", n->type);
	}
	return 0;
}


void free_node(node *n)
{
	if(--n->ref_cnt > 0)
		return;
	switch(n->type)
	{		
		case node_const:
		case node_variable:
			break;
		case node_binop:
			free_node(n->x.op.right);
		case node_unop:
			free_node(n->x.op.left);
			break;
		case node_assign:
			free_node(n->x.a.left);
			break;
		case node_selection:
		case node_rangecheck:
			free_node(n->x.sel.right);
			break;
		default:
			fatal_error("free_node: bad node type (%d)", n->type);
	}
	FREE(n);
}

typedef struct operator1
{
	int ch;
	int is_unary;
	binop_t opf;
	int prec;
} operator1;

typedef struct operator2
{
	int ch;
	operator1 *rest;
} operator2;

static operator1 sch_ops[] = {{'+', 0, binop_add, 3}, {'-', 0, binop_sub, 3}, 
							  {'*', 0, binop_mul, 2}, {'/', 0, binop_div, 2}, 
							  {'%', 0, binop_rem, 2},
							  {'^', 0, binop_xor, 4}, 
							  {'~', 1, unop_inv, 1},
							  {'=', 0, binop_eq, 5},
							  {',', 0, binop_seq, 10},
							  {0, 0, NULL, 0}
							 };
							 
static operator1 not_ops[2] = {{'=', 0, binop_neq, 5}, 
							   {0, 1, unop_lnot, 1}
							  };
							  
static operator1 and_ops[2]	= {{'&', 0, binop_land, 6}, 
							   {0, 0, binop_and, 4}
							  };
							  
static operator1 or_ops[2] = {{'|', 0, binop_lor, 6},
							  {0, 0, binop_or, 4}
						     };
							 
static operator1 left_ops[3] = {{'<', 0, binop_shiftl, 4},
								{'=', 0, binop_le, 5},
								{0, 0, binop_less, 5}
							   };		
							   
static operator1 right_ops[3] = {{'>', 0, binop_shiftr, 4},
								{'=', 0, binop_ge, 5},
								{0, 0, binop_greater, 5}
							   };		
							   
static operator2 dch_ops[] = {{'!', not_ops}, {'&', and_ops}, {'|', or_ops}, 
							  {'<', left_ops}, {'>', right_ops}, {0, NULL}};

// static operator1 paren = {{'(', 1, NULL, 1000}};

static operator1 op_neg = {'-', 1, unop_neg, 1};

enum opstack_types { st_operator, st_range, st_rangecheck, st_paren};
typedef struct opstack
{
	int type;
	union
	{
		operator1 *op;
		range *r;
/*		autoseq *seq; */
	} x;
} opstack;

static int range_len(range *r)
{
	int len = 0;
	for(; r->repeat > 0; r++)
		len++;
	return len;
}

range *copy_range(range *r)
{
	range *nr;
	int l;
	if(!r)
		return NULL;
	nr = xmalloc(l = sizeof(range) * (range_len(r) + 1));
	memcpy(nr, r, l);
	return nr;
}

static int range_find(rangedef *rd, char *name)
{
	return *rd->name && !strcmp(rd->name, name);
}


/// \section[RANGES]
/// (aka tables)
/// Ranges may have two syntactic forms: constant ranges and
/// range references.
//
/// Range references have the form \em[\[=range-name\]],
// where \em[range-name] is defined by a preceding 
// \see[directives][.range] directive
//
// Constant ranges are a comma-separated list of segments, enclosed
// in brackets.
// Each segment has the following structure:
// \em[start-end/dither*multi] 
// where every part except for \em[start] may be omitted: if \em[end]
// is omitted, it is implied equal to \em[start]; if 
// \em[dither] or \em[multi] is omitted, 1 is implied
// This means ``take \em[multi] times from \em[start] to \em[end]
// each \em[dither] element''
//
// \em2[start, end, dither, multi] should be integral or character constants;
// there should be no spaces around \em[-, /, *]
//
// To check whether a value is in a range, \em[@] operator is used. 
// It has the same precedence as unary operator; its left operand is
// a range, and the right is some expression. It returns a zero-base index of
// the first occurrence of a value within the range; or -1 if the value is outside
// the range.
//
// To select n-th item from a range, \em[application] operator is used;
// it has no symbolic representation; just write a range followed by an expression.
// The application has the same priority as \em[@].

#define RANGE_DELTA 16
range *parse_range(char *start, char **end, rangedef **rgs)
{
	int maxcount = RANGE_DELTA;
	range *r = xmalloc(maxcount * sizeof(range)), *nr = r;
	int count = 0;
	int base = 0;
	
	start++;
	if(*start == '=')
	{
		char *name = parse_plain_id(start + 1, &start);
		rangedef *rd;
		if(*start != ']')
			recover_error("missing or misplaced ']'");
		rd = list_find(*rgs, name, (list_predicate_t)range_find);
		if(!rd)
			recover_error("unknown range name: '%s'", name);
		if(end)
			*end = start + 1;
		return rd->r;
	}
	for(;;)
	{
		skip_spaces(start, &start);
		switch(*start)
		{
			case ',':
				start++;
				break;
			case '\0':
				recover_error("unmatched '[' occured");
			case ']':
			{
					rangedef *nrd;
					if(end)
						*end = start + 1;
					nrd = ALLOC(rangedef);
					nrd->name = xstrdup("");
					if(count == maxcount)
						r = xrealloc(r, sizeof(range) * (maxcount + 1));
					else
						r = xrealloc(r, sizeof(range) * (count + 1));
					r[count].repeat = -1;
					r[count].base = base;
					nrd->r = r;
					*rgs = list_add(*rgs, nrd);
					return r;
			}
			default:
				count++;
				if(count > maxcount)
				{
					maxcount += RANGE_DELTA;
					r = xrealloc(r, sizeof(range) * maxcount);
					nr = r + count - 1;
				}
				nr->base = base;
				nr->start = parse_num_token(start, &start);
				if(*start != '-')
					nr->end = nr->start;
				else nr->end = parse_num_token(start + 1, &start);
				if(*start != '/')
					nr->dither = 1;
				else nr->dither = strtoul(start + 1, &start, 0);
				if(*start != '*')
					nr->repeat = 1;
				else nr->repeat = strtoul(start + 1, &start, 0);
				if(nr->end < nr->start)
					nr->dither = -nr->dither;
				base += nr->end >= nr->start ? ((nr->end - nr->start + 1) / nr->dither) * nr->repeat :
					((nr->start - nr->end + 1) / -nr->dither) * nr->repeat;
				nr++;
		}
	}
}

static opstack stack[64], *sp;
static node *valstack[64], **vsp;

static void resolve_stack(int prec)
{
	for(;;)
	{
		if(sp == stack)
			recover_error("syntax error");
		switch(sp[-1].type)
		{
			case st_paren:
				return;
			case st_operator:
				if(sp[-1].x.op->prec > prec || (sp[-1].x.op->prec == prec && sp[-1].x.op->is_unary))
					return;	
				if(sp[-1].x.op->is_unary)
				{
					if(vsp == valstack)
						recover_error("missing argument of unary '%c'", sp[-1].x.op->ch);
					vsp[-1] = create_unop_node(sp[-1].x.op->opf, vsp[-1]);
				}
				else
				{
					if(vsp < valstack + 2)
						recover_error("missing argument(s) of binary '%c'", sp[-1].x.op->ch);
					vsp[-2] = create_binop_node(sp[-1].x.op->opf, vsp[-2], vsp[-1]);
					vsp--;
				}
				sp--;
				break;
			case st_range:
				if(vsp == valstack)
					recover_error("missing argument of selection");
				vsp[-1] = create_selection_node(sp[-1].x.r, vsp[-1]);
				sp--;
				break;
			case st_rangecheck:
				if(vsp == valstack)
					recover_error("missing argument of range-check");
				vsp[-1] = create_rangecheck_node(sp[-1].x.r, vsp[-1]);
				sp--;
				break;
			default:
				fatal_error("internal error");
		}
	}
}

static operator1 *find_single_op(int ch, operator1 *ops)
{
	for(; ops->ch; ops++)
	{
		if(ops->ch == ch)
			return ops;
	}
	return ops->opf ? ops : NULL;
}

static operator1 *find_leader(int ch, operator2 *ops)
{
	for(; ops->ch; ops++)
	{
		if(ops->ch == ch)
			return ops->rest;
	}
	return NULL;
}

operator1 *find_op(char *start, char **end)
{
	operator1 *op = find_single_op(*start, sch_ops);

	if(!op)
	{
		op = find_leader(*start, dch_ops);
		if(!op)
			recover_error("unknown character '%c'", *start);
		op = find_single_op(start[1], op);
		if(!op)
			recover_error("unknown character '%c'", start[1]);
		if(op->ch)
			start++;
	}
	*end = start + 1;
	return op;
}

static void destroy_macrodef(macrodef *dest)
{
	free(dest->name);
	free_node(dest->val);
}

static void destroy_rangedef(rangedef *dest)
{
	free(dest->name);
	free(dest->r);
}

static int find_macrodef(macrodef *dest, char *key)
{
	return !strcmp(dest->name, key);
}

node *parse_expr(char *start, char **end, macrodef *macros, rangedef **rgs)
{
	int is_prev_num = 0;
	sp = stack + 1;
	vsp = valstack;
	stack->type = st_paren;
	for(;;)
	{
		skip_spaces(start, &start);
		switch(*start)
		{
			case ';':
				start++;
			case '\0':
				if(end)
					*end = start;
				resolve_stack(INT_MAX);
				if(sp != stack + 1 || stack->type != st_paren)
					recover_error("ummatched parentheses");
			    if(vsp != valstack + 1)
					recover_error("syntax error");
				return use_node(*valstack);
			case '(':
				start++;
				sp->type = st_paren;
				sp++;
				is_prev_num = 0;
				break;
			case ')':
				start++;
				resolve_stack(INT_MAX);
				if(sp == stack || sp[-1].type != st_paren)
					recover_error("umatched parentheses");
				sp--;
				is_prev_num = 1;
				break;
			case '@':
				resolve_stack(1);
				sp->x.r = parse_range(start + 1, &start, rgs);
				sp->type = st_rangecheck;
				sp++;
				is_prev_num = 0;
				break;
			case ':':
				if(!isdigit(*++start))
					recover_error("invalid register: '%c'", *start);
				resolve_stack(1);
				if(vsp == valstack)
					recover_error("missing left part of ':'");
				vsp[-1] = create_assign_node(*start - '0' + nv_arg0, vsp[-1]);
				start++;
				is_prev_num = 0;
				break;
			case '[':
				sp->x.r = parse_range(start, &start, rgs);
				sp->type = st_range;
				sp++;
				is_prev_num = 0;
				break;
			case '$':
				switch(*++start)
				{
					case '$':
						*vsp++ = create_variable_node(nv_input);
						break;
					case '#':
						*vsp++ = create_variable_node(nv_depth);
						break;
					default:
						if(isdigit(*start))
							*vsp++ = create_variable_node(*start - '0' + nv_arg0);
						else
							recover_error("invalid character after $: '%c'", *start);				
				}
				start++;
				is_prev_num = 1;
				break;
			case '\'':
				*vsp++ = create_value_node(parse_char(start, &start));
				is_prev_num = 1;
				break;
			case '-':
				if(!is_prev_num)
				{
					resolve_stack(op_neg.prec);
					sp->x.op = &op_neg;
					sp->type = st_operator;
					sp++;
					is_prev_num = 0;
					start++;
					break;
				}
			default:
				if(isdigit(*start))
				{
					*vsp++ = create_value_node(strtoul(start, &start, 0));
					is_prev_num = 1;
				}
				else if(isalpha(*start))
				{
					char buffer[64];
					char *bufptr = buffer;
					macrodef *md;

					while(isnamesym(*start))
						*bufptr++ = *start++;
					*bufptr = '\0';
					md = list_find(macros, buffer, (list_predicate_t)find_macrodef);
					if(!md)
						recover_error("undefined macro %s", buffer);
					*vsp++ =  md->val;
					is_prev_num = 1;
				}
				else
				{
					operator1 *op = find_op(start, &start);
					resolve_stack(op->prec);
					sp->x.op = op;
					sp->type = st_operator;
					sp++;
					is_prev_num = 0;
				}
		}
	}
}

int eval_expr(char *start, char **end, macrodef *macros, rangedef *rgs)
{
	rangedef *ords = rgs;
	node *n = parse_expr(start, end, macros, &rgs);
	static reg_block_t dummy;
	int v = eval_node(n, dummy);
	free_node(n);
	while(rgs != ords)
		rgs = list_pop(rgs, (list_destroy_t)destroy_rangedef, sizeof(rangedef));
	return v;
}

#define AUS_STACK_INCR 10

static void push_to_autostack(auto_state *as, int value)
{
	if(as->depth == as->maxdepth)
	{
		as->maxdepth += AUS_STACK_INCR;
		as->stack = xrealloc(as->stack, as->maxdepth * sizeof(*as->stack));
	}
	as->stack[as->depth++] = value;
}

static int find_transition(transition *trans, auto_state *as)
{
	return eval_node(trans->filter, as->regs) && 
		   eval_node(trans->stackfilter, as->regs);
}


static int change_autostate(auto_state *as, int *inchar)
{
	node **iter;
	int s;

	if(as->depth)
		as->depth--;
	if(as->active->tostack)
	{
		for(iter = as->active->tostack; *iter; iter++)
		{
			if((*iter)->type == node_assign)
				eval_node(*iter, as->regs);
			else
				push_to_autostack(as, eval_node(*iter, as->regs));
		}
	}
	s = eval_node(as->active->newstate, as->regs);
	if(s < 0 || s >= as->au->page_cnt * (1 << STATE_PAGE_BITS))
		recover_error("invalid state: %d", s);
	as->current = s;
	if(as->active->eat_input)
		*inchar = -1;
	as->active = NULL;
	return ar_accepted;
}

/// \section[REGISTERS]
// The memory of an automaton consists of a stack, a current state, an input characters and 8 temporary
// registers. They are accessed in the following way:
// \item $$ -- an input character
// \item $# -- stack depth
// \item $0 -- current state
// \item $1 -- top of the stack or 0 if the stack is empty
// \item $2..$9 -- temporary registers
// \endlist
//
// The registers may be assigned to with a \em[:] operator; its left arg is an expression
// and its right args is a register number. It has the same precedence as \em[@]
// \note 0 and 1 regs may be assigned, but this would \em[not] really change the state or the stack;
// they are just shadows of the actual values.
// \note temporary registers (2..9) are persisent in the sense that they \em[do] keep
// their value across transitions


int auto_step(auto_state *as, int *inchar, int *outchar)
{
	if(as->active)
	{
		*outchar = eval_node(*as->active_out, as->regs);
		as->active_out++;
		if(*as->active_out)
			return ar_accepted;
		return change_autostate(as, inchar);
	}
	else
	{
		transition *newt;
		as->regs[nv_input] = *inchar;
		as->regs[nv_state] = as->current;
		as->regs[nv_depth] = as->depth;
		as->regs[nv_topstack] = as->depth ? as->stack[as->depth - 1] : 0;
		newt = list_find(as->au->table[as->current >> STATE_PAGE_BITS][as->current & ((1 << STATE_PAGE_BITS) - 1)], 
						 as, (list_predicate_t)find_transition);
		as->active = newt;
		if(!newt)
		{
			as->accepted = 0;
			return ar_rejected;
		}
		if(!newt->outchars)
		{
			*outchar = -1;
			return change_autostate(as, inchar);
		}
		else
		{
			*outchar = eval_node(*newt->outchars, as->regs);
			if(newt->outchars[1])
			{
				as->active_out = newt->outchars + 1;
				return ar_accepted;
			}
			return change_autostate(as, inchar);
		}
	}
}

//// \section[TRANSITIONS]
//// Each automaton is essentially a set of transitions.
//// A transition definition has the following syntax:
//// \indent[transition ::=] state-expr `;' eat filter-expr1 `;' filter-expr2 `;' newstate-expr `;'
//// `{' expr-list1 `}' `{' expr-list2 `}'
/// \indent[eat ::=] \em[null] \par `^'
/// \indent[expr-list ::=] \em[null]
/// \par expr-or-string expr-list
/// \indent[expr-or-string ::=] expr `;' \par quoted-string
//// \noindent
//// 
//// A \em[state-expr] is evaluated at parse time yielding the number of a state.
//// It must non-negative.
///
/// If \em[eat] is `^', an input character of the automaton will not change after transition.
/// If \em[filter-exprs] are always true, this is equivalent to empty-string transition.
////
//// \em[Filter-exprs] are evaluated (at run time) in a short-circuit manner as if connected
/// by `&&' operator. If they're both true, a transition for a given state is executed.
/// \note Transitions are searched in the order of their definitions. After a matching transition
/// is found, search is complete. \noindent
//// By convention, \em[filter-expr1] shall check conditions related to input character, while
/// \em[filter-expr2] -- those related to stack and temporary registers. This is not obligatory,
/// however.
////
//// When a transition is executed, expressions from \em[expr-list1] are evaluated and their
//// values are passed to output. If a list item is a string, it is considered equivalent to
/// a list of character constants.
/// 
//// Then a topmost stack item is eliminated, and \em[expr-list2] is evaluated.
//// \note Here string items are, too, equivalent to sequence of characters. As a consequence,
/// a string will be pushed into the stack reversed.
//// \note If an expression from \var[expr-list2] is a standalone assignment, its result
//// is \em[not] pushed into the stack \noindent
////
//// After all, the state is changed to the value of \em[newstate-expr].
//// \note Topmost stack element and the stack depth are cached in `$1' and '$#', 
/// so when executing \em[expr-list2],
/// `$1' and `$#' will hold value the topmost stack element and stack depth 
/// had at the beginning of a transition, not the actual values. 


static void add_transition(automaton *au, transition *t)
{
	int pageno = t->state >> STATE_PAGE_BITS, offset = t->state & ((1 << STATE_PAGE_BITS) - 1);
	if(pageno >= au->page_cnt)
	{
		int i;
		au->table = xrealloc(au->table, (pageno + 1) * sizeof(*au->table));
		for(i = au->page_cnt; i <= pageno; i++)
			au->table[i] = NULL;
		au->page_cnt = pageno + 1;
	}
	if(!au->table[pageno])
		au->table[pageno] = xcalloc(1 << STATE_PAGE_BITS, sizeof(**au->table));
	au->table[pageno][offset] = list_append(au->table[pageno][offset], t);
	au->ditto = t;
}

typedef void (*ad_directive_t)(automaton *au, char *start, char **end);


//// \section[DIRECTIVES]
//// \define[directive][\if[#2][]\subsection[#1]\else\subsection[#1 #2]\endif\*[directives][#1]][2]

//// \directive[.define][name expression;]
//// defines a macro \em[name] which is expanded to \em[expression]
//// \note no evaluation occur at the point of definition
//// \note macros are expanded as if their body is surrounded by parentheses.
static void ad_define(automaton *au, char *start, char **end)
{
	macrodef *newm = ALLOC(macrodef);
	newm->name = xstrdup(parse_plain_id(start, &start));
	newm->val = use_node(parse_expr(start, end, au->names, &au->ranges));
	au->names = list_add(au->names, newm);
}

//// \directive[.include]["filename"]
//// includes a file \em[filename]
/// \note \em[filename] may be in fact surround by arbitrary characters,
/// but they need be the same -- \em[.include <...>] is not allowed.
static void ad_include(automaton *au, char *start, char **end)
{
	char *buf = load_file(parse_string(start, end), getenv(AU_PATH_ENV), "aut");
	int lno = set_line_no(1);
	parse_automaton(au, buf, NULL);
	free(buf);
	set_line_no(lno);
}

/// \directive[.start][expression;]
//// declares the state whose number is the value of \em[expression]
//// as a default start state
static void ad_start(automaton *au, char *start, char **end)
{
	au->start = eval_expr(start, end, au->names, au->ranges);
}

//// \directive[.final][expression;]
//// declares the state whose number is the value od \em[expression]
//// as final.
static void ad_final(automaton *au, char *start, char **end)
{
	int f = eval_expr(start, end, au->names, au->ranges);
	au->finals = xrealloc(au->finals, ++au->fin_cnt * sizeof(*au->finals));
	au->finals[au->fin_cnt - 1] = f;
}

//// \directive[.if][(.ifndef) name ... .else ... .endif]
//// Conditionally processes a piece of file depending on
/// whether a macro \em[name] is defined
static void ad_ifdef(automaton *au, char *start, char **end)
{
	if(list_find(au->names, parse_plain_id(start, end), (list_predicate_t)find_macrodef))
		return;
	skip_ifs(start, end, 1);
}

static void ad_ifndef(automaton *au, char *start, char **end)
{
	if(!list_find(au->names, parse_plain_id(start, end), (list_predicate_t)find_macrodef))
		return;
	skip_ifs(start, end, 1);
}

static void ad_else(automaton *au, char *start, char **end)
{
	skip_ifs(start, end, 0);
}

static void ad_endif(automaton *au, char *start, char **end)
{
	*end = start;
}

//// \directive[.ditto][expression;]
//// copies the most recently defined transition to a state with the number
/// which is equal to \em[expression]
/// \bug implementation is broken
static void ad_ditto(automaton *au, char *start, char **end)
{
	fatal_error(".ditto is broken");
/*
	transition *newt = ALLOC(transition);
	if(!au->ditto)
		recover_error("no previous transition for '.ditto'");
	newt = au->ditto; // very dangerous! fix ASAP!
	newt->state = eval_expr(start, end, au->names, au->ranges);
	if(newt->state < 0)
		recover_error("invalid state: %d", newt->state);
	add_transition(au, newt);
*/
}

//// \directive[.range][name range-constant\[;\]]
//// defines a range \em[name] whose value is \em[range-constant]
static void ad_range(automaton *au, char *start, char **end)
{
	char *name = xstrdup(parse_plain_id(start, &start));
	skip_spaces(start, &start);
	if(*start != '[')
		recover_error("syntax error");
	parse_range(start, &start, &au->ranges);
	skip_spaces(start, &start);
	if(*start == ';')
		start++;
	if(end)
		*end = start;
	free(au->ranges->name);
	au->ranges->name = name;
}

//// \directive[.end]
//// stops processing the current file.
/// Mostly useful for automata definitions embedded into Tensile programs.
static void ad_end(automaton *au, char *start, char **end)
{
	*end = start;
}

static int process_directives(automaton *au, char *start, char **end)
{
	static table_t directives[] = {{"define", ad_define}, {"include", ad_include}, 
								   {"start", ad_start}, {"final", ad_final},
								   {"ifdef", ad_ifdef}, {"ifndef", ad_ifndef},
								   {"else", ad_else}, {"endif", ad_endif},
								   {"ditto", ad_ditto},
								   {"range", ad_range},
								   {"end", ad_end},
								   {NULL, NULL}
								  };
							
	ad_directive_t d;
	char *id;
	
	start++;
	d = lookup_table(directives, id = parse_plain_id(start, &start));
	skip_spaces(start, &start);
	if(!d)
		recover_error("unknown directive: .%s", id);
	d(au, start, end);
	return d != ad_end;
}

node **parse_nodelist(char *start, char **end, macrodef *macros, rangedef **rgs)
{
	node **nodes = NULL;
	int count = 0;

	skip_spaces(start, &start);
	if(*start != '{')
		recover_error("'{' expected");
	start++;
	for(;;)
	{
		skip_spaces(start, &start);
		if(!*start)
			recover_error("missing '}'");
		if(*start == '}')
		{
			start++;
			if(end)
				*end = start;
			return nodes;
		}
		if(*start == '\"')
		{
			char *str = parse_string(start, &start);
			for(; *str; str++)
			{
				count++;
				nodes = xrealloc(nodes, sizeof(*nodes) * (count + 1));
				nodes[count - 1] = create_value_node(*str);
				nodes[count] = NULL;
			}
		}
		else
		{
			count++;
			nodes = xrealloc(nodes, sizeof(*nodes) * (count + 1));
			nodes[count - 1] = parse_expr(start, &start, macros, rgs);
			nodes[count] = NULL;
		}
	}
}

void parse_automaton(automaton *au, char *start, char **end)
{
	for(;;)
	{
		skip_spaces(start, &start);
		if(!*start)
			break;
		if(*start == '.')
		{
			if(!process_directives(au, start, &start))
				break;
		}
		else
		{
			transition *newt = ALLOC(transition);
			newt->state = eval_expr(start, &start, au->names, au->ranges);
			skip_spaces(start, &start);
			if(*start != '^')
				newt->eat_input = 1;
			else
			{
				start++;
				newt->eat_input = 0;
			}
			newt->filter = parse_expr(start, &start, au->names, &au->ranges);
			newt->stackfilter = parse_expr(start, &start, au->names, &au->ranges);
			newt->newstate = parse_expr(start, &start, au->names, &au->ranges);
			newt->outchars = parse_nodelist(start, &start, au->names, &au->ranges);
			newt->tostack  = parse_nodelist(start, &start, au->names, &au->ranges);
			skip_spaces(start, &start);
			add_transition(au, newt);
		}
	}
	if(end)
		*end = start;
}


static automaton *loaded_aus;

static int find_automaton(automaton *au, char *name)
{
	return !strcmp(au->name, name);
}

automaton *get_au_byname(char *name)
{
	return list_find(loaded_aus, name, (list_predicate_t)find_automaton);
}

automaton *create_automaton(char *name)
{
	automaton *au = list_find(loaded_aus, name, (list_predicate_t)find_automaton);
	if(au)
		recover_error("automaton '%s' already defined", name);
	au = ALLOC(automaton);
	au->name = xstrdup(name);
	loaded_aus = list_add(loaded_aus, au);
	return au;
}

automaton *load_automaton(char *name)
{
	char *buf;
	int old_line_no = set_line_no(1);
	automaton *au = list_find(loaded_aus, name, (list_predicate_t)find_automaton);
		
	if(au)
		return au;
	au = ALLOC(automaton);
	buf = load_file(name, getenv(AU_PATH_ENV), "aut");
	au->name = xstrdup(name);
	au->names = NULL;
	au->table = NULL;
	au->finals = NULL;
	au->fin_cnt = 0;
	parse_automaton(au, buf, NULL);
	loaded_aus = list_add(loaded_aus, au);
	free(buf);
	set_line_no(old_line_no);
	return au;
}

int map_auto_name(automaton *au, char *name)
{
	macrodef *md = list_find(au->names, name, (list_predicate_t)find_macrodef);
	static reg_block_t dummy;
	if(!md)
	{
		errno = ENOENT;
		return -1;
	}
	return eval_node(md->val, dummy);
}

static int state_id;
static int find_macrobyid(macrodef *m, char *prefix)
{
	static reg_block_t dummy;
	if(!strncmp(m->name, prefix, strlen(prefix)))
		return eval_node(m->val, dummy) == state_id;
	return 0;
}

char *get_auname_byid(automaton *au, int id, char *prefix)
{
	macrodef *md;
	state_id = id;
	md = list_find(au->names, prefix, (list_predicate_t)find_macrobyid);
	if(!md)
	{
		errno = ENOENT;
		return NULL;
	}
	return md->name;
}

int is_in_final(auto_state *as)
{
	int *ptr, i;
	for(i = as->au->fin_cnt, ptr = as->au->finals; i; i--, ptr++)
	{
		if(*ptr == as->current)
			return 1;
	}
	return 0;
}

void init_auto_state(auto_state *as, automaton *au)
{
	as->au = au;
	as->current = au->start;
	as->maxdepth = AU_START_STACK_DEPTH;
	as->depth = 0;
	as->active = NULL;
	as->active_out = NULL;
	as->stack = xcalloc(as->maxdepth, sizeof(*as->stack));
	as->accepted = 1;
}

void cleanup_auto_state(auto_state *as)
{
	as->active = NULL;
	as->active_out = NULL;
	as->depth = 0;
	as->accepted = 1;
}

static au_ext_create_f ext_creator;
static au_ext_process_f ext_processor;
static au_ext_service_f ext_servicer;
void register_au_extension(au_ext_create_f on_create, au_ext_process_f on_process,
						   au_ext_service_f on_service)
{
	if(on_create)
		ext_creator = on_create;
	if(on_process)
		ext_processor = on_process;
	if(on_service)
		ext_servicer = on_service;
}

//// \section[AUTOMATA SEQUENCES]
//// \subsection[Syntax]
//// Automata sequences has the following syntax:
/// \indent[autoseq ::=] \em[null]
/// \par branch
/// \par branch `|' autoseq
/// \indent[branch ::=] item
/// \par item `&' branch
/// \indent[item ::=] ismaster automaton param-spec outfunc
/// \indent[ismaster ::=] \em[null]
/// \par `*'
/// \indent[automaton ::=] auto-name state-spec
/// \par `{' extension-name extension-param '}'
/// \indent[param-spec ::=] \em[null]
/// \par `[' params `]'
/// \indent[params ::= ] param \par param `,' params
/// \indent[param ::= ] reg-spec `=' number-or-name
/// \indent[reg-spec ::= ] digit \par name
/// \indent[number-or-name ::=] number \par char-constant \par name
/// \indent[auto-name ::=] name \par apos-string
/// \indent[state-spec ::=] \em[null]
//// \par `:' name
/// \indent[extension-name ::=] \em[any characters except \] and =]
/// \indent[extension-param ::=] \em[null] \par \em[any characters except \]]
/// \indent[outfunc ::=] \em[null] \par `$$' \par `$1' \par `$0' \par `$#'
/// \noindent
///
/// Arbitrary spaces are allowed between tokens
autoseq *parse_autoseq(char *start)
{
	autoseq *seq = NULL;
	autoseq *ns;
	autoseq_branch *br;
	automaton *au;
	char *buf;
	char *id;
	int has_master = 0;
	int no;
	macrodef *md;
	
	for(;;)
	{
		for(;;)
		{
			skip_spaces(start, &start);
			if(*start != '|')
				break;
			start++;
		}
		if(!*start)
		{
			if(seq && !has_master)
				seq->branches->master = 1;
			return seq;
		}
		ns = ALLOC(autoseq);
		for(;;)
		{
			br = ALLOC(autoseq_branch);
			if(*start == '*')
			{
				if(has_master)
					recover_error("multiple masters specified");
				br->master = 1;
				has_master = 1;
				start++;
			}
			if(*start == '{')
			{
				char *parm = NULL;
				if(!ext_creator)
					recover_error("no automaton extension registered");
				id = parse_id(start + 1, &start);
				if(*start == '=')
				{
					parm = ++start;
					start = strchr(start, '}');
				}
				if(parm)
				{
					parm = xstrdup(parm);
					*strchr(parm, '}') = '\0';
				}
				if(!start || *start != '}')
				{
					free(id);
					recover_error("syntax error at extension");
				}
				start++;
				br->extension = ext_creator(id, parm);
				if(!br->extension)
					recover_error("can't create automaton extension");
				free(parm);
			}
			else
			{
				id = *start == '\'' ? parse_string(start, &start) : parse_plain_id(start, &start);
				init_auto_state(&br->state, au = load_automaton(id));
				skip_spaces(start, &start);
				if(*start != ':')
					br->initial[nv_state] = br->state.au->start;
				else
				{
					start++;
					if(*start == '*')
					{
						br->initial[nv_state] = -1;
						start++;
					}
					else
					{
						errno = 0;
						br->initial[nv_state] = map_auto_name(au, buf = parse_plain_id(start, &start));
						if(errno)
							recover_error("name '%s' is not defined in '%s'", buf, br->state.au->name);
					}
				}
				br->initial[nv_depth] = 0;
				if(*start == '[')
				{
					start++;
					for(;;)
					{
						if(*start > '0' && *start <= '9')
							no = *start++ - '0';
						else
						{
							id = parse_plain_id(start, &start);
							md = list_find(au->names, id, (list_predicate_t)find_macrodef);
							if(!md || md->val->type != node_variable)
								recover_error("invalid autoseq macro: %s", id);
							no = md->val->x.val;
						}
						if(no == nv_topstack)
							br->initial[nv_depth] = 1;
						if(*start != '=')
							recover_error("syntax error in autoseq");
						start++;
						if(isdigit(*start) || *start == '\'')
							br->initial[no] = parse_num_token(start, &start);
						else
						{
							errno = 0;
							br->initial[no] = map_auto_name(au, id = parse_plain_id(start, &start));
							if(errno)
								recover_error("unknown macro name %s", id);
						}
						if(*start == ']')
							break;
						if(*start != ',')
							recover_error("\',\' expected");
						skip_spaces(start + 1, &start);
					}
					start++;
				}
			}
			if(*start == '$')
			{
				switch(*++start)
				{
					case '$':
						break;
					case '0':
						br->outfunc = aof_state;
						break;
					case '1':
						br->outfunc = aof_topstack;
						break;
					case '#':
						br->outfunc = aof_depth;
						break;
					default:
						recover_error("unknown output function: $%c", *start);
				}
				start++;
			}
			ns->branches = list_add(ns->branches, br);
			if(*start != '&')
				break;
			start++;
		}
		ns->branches = list_reversip(ns->branches);
		seq = list_add(seq, ns);
	}
}

static void ab_init(autoseq_branch *ab, void *unused)
{
	int i;
	ab->state.accepted = 1;
	if(ab->extension)
	{
		ext_servicer(ab->extension, aeo_reset, 0, NULL);
		return;
	}
	if(ab->initial[nv_state] >= 0)
		ab->state.current = ab->initial[nv_state];
	ab->state.depth = 0;
	if(ab->initial[nv_depth])
		push_to_autostack(&ab->state, ab->initial[nv_topstack]);
	for(i = nv_arg2; i <= nv_last; i++)
		ab->state.regs[i] = ab->initial[i];
	ab->state.active = NULL;
}

static void as_init(autoseq *seq, void *unused)
{
	seq->inch = -1;
	list_foreach(seq->branches, (list_iterator_t)ab_init, NULL);
}

void autoseq_init(autoseq *seq)
{
	list_foreach(seq, (list_iterator_t)as_init, NULL);
}

//// \subsection[Evaluation]
/// A sequence is evaluated right to left, the input of an automata branch being
/// the output of the next one (in evaluation order).
/// 
/// A branch is evaluated left to right, the output of the branch being the first
/// non-empty output. However, all the branches are always executed.
////
//// An output value of an automaton can be modified by an \em[output function]
/// which now can substitute a current state, a topmost stack element or a stack
/// depth for a real output value.
////
//// When a sequence is initialized or reset, all its automata states are set to
/// the value of \em[state-spec] or to a value defined by \see[directives][.start] 
/// if the former is missing.
//// Temporary registers and stack may be initialized by additional \em[params].
//// A \em[param] designator may be either a digit or a macro name which expands to
/// a register reference. A \em[param] value may be either a number or a character
/// constant, or a macro name which is get evaluated at parse time.
////
/// By default, when polling a state of a sequence, the last branch of the last
/// sequence item (i. e. the rightmost one) is used. This can be changed by
/// prefixing a given item with `*'.
int autoseq_step(autoseq *seq, int *inchar, int *outchar)
{
	int result = ar_accepted;
	autoseq_branch *bi;
	int tmp, out, in = -1, r = ar_accepted;
	
	if(!seq)
	{
		*outchar = *inchar;
		*inchar = -1;
		return ar_accepted;
	}
	if(seq->inch < 0)
	{
		result = autoseq_step(list_next(seq), inchar, &seq->inch);
		if(result == ar_rejected)
		{
			for(bi = seq->branches; bi; bi = list_next(bi))
				bi->state.accepted = 0;
			return ar_rejected;
		}
		if(seq->inch < 0)
		{
			*outchar = seq->inch;
			return result;
		}
	}
	*outchar = -1;
	for(bi = seq->branches; bi; bi = list_next(bi))
	{
		tmp = seq->inch; 
		r = bi->extension ? ext_processor(bi->extension, &tmp, &out) : 
			auto_step(&bi->state, &tmp, &out);
		switch(bi->outfunc)
		{
			case aof_state:
				out = bi->extension ? ext_servicer(bi->extension, aeo_getreg, nv_state, NULL) : 
					bi->state.current;
				break;
			case aof_topstack:
				out = bi->extension ? ext_servicer(bi->extension, aeo_getreg, nv_topstack, NULL) : 
					(bi->state.depth ? bi->state.stack[bi->state.depth - 1] : 0);
				break;
			case aof_depth:
				out = bi->extension ? ext_servicer(bi->extension, aeo_getreg, nv_depth, NULL) :
					bi->state.depth;
				break;
		}
		if(out >= 0 && *outchar < 0)
		{
			*outchar = out;
			in = tmp;
			result = r;
		}
	}
	seq->inch = *outchar >= 0 ? in : tmp;
	return *outchar >= 0 ? result : r;
}

int need_feeding(autoseq *seq)
{
	if(!seq)
		return 1;
	return seq->inch < 0 ? need_feeding(list_next(seq)) : 0;
}

int as_get_reg(autoseq *seq, int reg)
{
	autoseq_branch *bi;
	if(!seq)
		return 0;
	for(bi = seq->branches; bi; bi = list_next(bi))
	{
		if(bi->master)
		{
			if(bi->extension)
				return ext_servicer(bi->extension, aeo_getreg, reg, NULL);
			switch(reg)
			{
				case nv_state:
					return bi->state.current;
				case nv_topstack:
					return bi->state.depth ? bi->state.stack[bi->state.depth - 1] : 0;
				case nv_depth:
					return bi->state.depth;
				default:
					return reg >= nv_arg2 && reg <= nv_last ? bi->state.regs[reg] : 0;
			}
		}
	}
	return as_get_reg(list_next(seq), reg);
}

int as_map_name(autoseq *seq, char *name)
{
	autoseq_branch *bi;
	if(!seq)
	{
		errno = ENOENT;
		return -1;
	}
	for(bi = seq->branches; bi; bi = list_next(bi))
	{
		if(bi->master)
			return bi->extension ? ext_servicer(bi->extension, aeo_mapname, 0, name) : 
				map_auto_name(bi->state.au, name);
	}
	return as_map_name(list_next(seq), name);
} 

char *as_getname_byid(autoseq *seq, int id, char *prefix)
{
	autoseq_branch *bi;
	if(!seq)
		return NULL;
	for(bi = seq->branches; bi; bi = list_next(bi))
	{
		if(bi->master)
		{
			if(bi->extension)
				recover_error("not implemented");
			return get_auname_byid(bi->state.au, id, prefix);
		}
	}
	return as_getname_byid(list_next(seq), id, prefix);
}

int as_is_in_final(autoseq *seq)
{
	autoseq_branch *bi;
	if(!seq)
		return 1;
	for(bi = seq->branches; bi; bi = list_next(bi))
	{
		if(bi->master)
			return bi->extension ? ext_servicer(bi->extension, aeo_isfinal, 0, NULL) : 
				is_in_final(&bi->state);
	}
	return as_is_in_final(list_next(seq));
}

int as_is_accepted(autoseq *seq)
{
	autoseq_branch *bi;
	if(!seq)
		return 1;
	for(bi = seq->branches; bi; bi = list_next(bi))
	{
		if(bi->master)
			return bi->state.accepted;
	}
	return as_is_accepted(list_next(seq));
}

static void destroy_autostate(autoseq_branch *ab)
{
	if(ab->extension)
		ext_servicer(ab->extension, aeo_destroy, 0, NULL);
	else
		free(ab->state.stack);
}

static void destroy_autoseq(autoseq *as)
{
	list_delete(as->branches, (list_destroy_t)destroy_autostate, sizeof(*as->branches));
}

void remove_autoseq(autoseq *as)
{
	list_delete(as, (list_destroy_t)destroy_autoseq, sizeof(*as));
}

autoseq *clone_autoseq(autoseq *src)
{
	autoseq *clone = NULL, *na;
	autoseq_branch *nb, *iter;
	int i;

	for(; src; src = list_next(src))
	{
		na = ALLOC(autoseq);
		na->inch = src->inch;
		for(iter = src->branches; iter; iter = list_next(iter))
		{
			if(iter->extension)
				recover_error("not implemented");
			nb = ALLOC(autoseq_branch);
			nb->outfunc = iter->outfunc;
			nb->master  = iter->master;
			for(i = 0; i  < NUMOFITEMS(iter->initial); i++)
			{
				nb->initial[i] = iter->initial[i];
				nb->state.regs[i] = iter->state.regs[i];
			}
			nb->state.au = iter->state.au;
			nb->state.accepted = iter->state.accepted;
			nb->state.current = iter->state.current;
			nb->state.depth = iter->state.depth;
			nb->state.maxdepth = iter->state.maxdepth;
			nb->state.stack = xmalloc(nb->state.maxdepth * sizeof(*nb->state.stack));
			memcpy(nb->state.stack, iter->state.stack, nb->state.maxdepth * sizeof(*nb->state.stack));
			nb->state.active = iter->state.active;
			nb->state.active_out = iter->state.active_out;
			na->branches = list_add(na->branches, nb);
		}
		na->branches = list_reversip(na->branches);
		clone = list_add(clone, na);
	}
	return list_reversip(clone);
}
