/*
 	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <modules.h>
#include <interp.h>
#include <util.h>
#include <plist.h>
#include <dstring.h>
#include <fflow.h>
#include "version.h"
//// \author[Artem V. Andreev]
//// \copyright[2001, 2002 Artem V. Andreev -- See \ref[sl(1)] for details]

#ifndef HAVE_SOCKLEN_T
typedef int socklen_t;
#endif

/// \section[NAME]
// \em[sockets] -- a module for Tensile (see \ref[sl(1)])

/// \section[OVERVIEW]
//// \em[sockets] provides access to TCP connections.
//// \todo[UNIX sockets]
//// \todo[listening for connections]
//// \todo[proxies support]

static module *this_mod;
extern procstack *get_pstack(void);


static struct in_addr make_in_addr(char *str)
{
	struct in_addr ia;
	if(!inet_aton(str, &ia))
	{
		struct hostent *he = gethostbyname(str);
		if(!he)
			recover_error("unknown host '%s': %d", str, h_errno);
		if(he->h_addrtype != AF_INET)
			recover_error("host '%s' has non-Internet address", str);
		ia = *(struct in_addr *)he->h_addr;
	}
	return ia;
}

static struct sockaddr_in *make_sock_addr(char *str)
{
	static struct sockaddr_in sa;
	char *port, *tmp;
	strtok(str, ":");
	port = strtok(NULL, ":");
	if(!port)
		recover_error("'%s': no port name given", str);
	sa.sin_family = AF_INET;
	sa.sin_addr = make_in_addr(str);
	sa.sin_port = htons(strtoul(port, &tmp, 10));
	if(*tmp)	
	{
		struct servent *se = getservbyname(port, "tcp");
		if(!se)
			recover_error("unknown service: '%s'", port);
		sa.sin_port = se->s_port;
	}
	return &sa;
}


static void hostinfo_destroy(interpreter *i, storage *s)
{
	struct hostent *host = s->data;
	char **iter;
	free(host->h_name);
	for(iter = host->h_aliases; *iter; iter++)
		free(*iter);
	free(host->h_aliases);
	for(iter = host->h_addr_list; *iter; iter++)
		free(*iter);
	FREE(host);
}

//// \section[HOSTINFO STORAGE]
static char *hostinfo_fields[] = {"name", "aliases", "address", "addresses", NULL};

static int hostinfo_check(interpreter *i, storage *s, dstring *key)
{
	return lookup_name(hostinfo_fields, DS_BODY(key)) >= 0;
}

static dstring *hostinfo_get(interpreter *i, storage *s, dstring *key)
{
	struct hostent *he = s->data;
	storage *lst;
	int j = lookup_name(hostinfo_fields, DS_BODY(key));
	char *name = NULL;
	char **iter;
	dstring *buffer = ds_create(NULL);
	dstring *null = ds_create(NULL);
	
	if(j < 0)
		recover_error("hostinfo: unknown field '%s'", DS_BODY(key));
	switch(j)
	{
		case 0:
			return ds_create(he->h_name);
		case 1:
			lst = create_storagex(i, &plist_storage_class, &name, NULL);
			for(iter = he->h_aliases; *iter; iter++)
			{
				ds_appendstr(buffer, *iter);
				lst->class->add(i, lst, buffer, null);
				ds_assign(buffer, NULL);
			}
			return ds_create(name);
		case 2:
			return ds_create(inet_ntoa(*(struct in_addr *)he->h_addr));
		case 3:
			lst = create_storagex(i, &plist_storage_class, &name, NULL);
			for(iter = he->h_addr_list; *iter; iter++)
			{
				ds_appendstr(buffer, inet_ntoa(*(struct in_addr *)he->h_addr));
				lst->class->add(i, lst, buffer, null);
				ds_assign(buffer, NULL);
			}
			return ds_create(name);
	}
	fatal_error("hostinfo_get: internal error");
	return NULL; // never reached
}

static void hostinfo_dummy1(interpreter *i, storage *s, dstring *key, dstring *val)
{
	recover_error("hostinfo cannot be modified");
}

static void hostinfo_dummy2(interpreter *i, storage *s, dstring *key)
{
	recover_error("hostinfo cannot be modified");
}

static dstring *hostinfo_first(interpreter *i, storage *s, void **iter)
{
	*(int *)iter = 1;
	return ds_create("name");
}

static dstring *hostinfo_next(interpreter *i, storage *s, void **iter)
{
	dstring *tmp;
	if(!hostinfo_fields[*(int *)iter])
		return NULL;
	tmp = ds_create(hostinfo_fields[*(int *)iter]);
	(*(int *)iter)++;
	return tmp;
}

static void hostinfo_enditer(interpreter *i, storage *s, void **iter)
{
}

static int hostinfo_numofitems(interpreter *i, storage *s)
{
	return 4;
}

static storage *hostinfo_create(interpreter *i, char *name, char *arg);
static storage_class hostinfo_storage_class = {hostinfo_create, hostinfo_destroy, 
									 hostinfo_check, hostinfo_get,
									 hostinfo_dummy1, hostinfo_dummy1,
									 hostinfo_dummy2,
									 hostinfo_first, hostinfo_next,
									 hostinfo_enditer,
									 hostinfo_numofitems
									};


static storage *hostinfo_create(interpreter *i, char *name, char *arg)
{
	storage *ns = ALLOC(storage);
	struct in_addr ia = make_in_addr(arg);
	struct hostent *host = gethostbyaddr(&ia, sizeof(ia), AF_INET), *nh;
	char **iter, **iter1;
	
	if(!host)
		recover_error("can't get host info for '%s': %d", arg, h_errno);
	ns->class = &hostinfo_storage_class;
	if(host->h_addrtype != AF_INET)
		recover_error("host '%s' has non-Internet address", arg);
	ns->data = nh = ALLOC(struct hostent);
	nh->h_name = xstrdup(host->h_name);
	for(iter = host->h_aliases; *iter; iter++)
		;
	nh->h_aliases = xcalloc(iter - host->h_aliases + 1, sizeof(*nh->h_aliases));
	for(iter = host->h_aliases, iter1 = nh->h_aliases; *iter; iter++, iter1++)
		*iter1 = xstrdup(*iter);
	for(iter = host->h_addr_list; *iter; iter++)
		;
	nh->h_addr_list = xcalloc(iter - host->h_addr_list + 1, sizeof(*nh->h_addr_list));
	for(iter = host->h_addr_list, iter1 = nh->h_addr_list; *iter; iter++, iter1++)
	{
		*iter1 = xmalloc(host->h_length);
		memcpy(*iter1, *iter, host->h_length);
	}
	nh->h_length = host->h_length;
	nh->h_addr = *nh->h_addr_list;
	return ns;
}

static flow_handle_t skf_fopen(char *name, char *options, int writeable)
{
	file_flow_data *fd = ALLOC(file_flow_data);
	int skt;
	fd->writeable = 1;
	fd->preload   = 0;
	fd->undone 	  = FLOW_EOF;
	skt = socket(PF_INET, SOCK_STREAM, 0);
	if(skt < 0)
		recover_error("can't open socket '%s': %s", name, strerror(errno));
	if(strstr(options, "reuseaddr"))
	{
		const int yes = 1;
		setsockopt(skt, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
	}
	if(connect(skt, make_sock_addr(name), sizeof(struct sockaddr_in)))
		recover_error("binding error on socket '%s': %s", name, strerror(errno));
	fd->x.hfile = fdopen(skt, "r+");
	if(!fd->x.hfile)
	{
		close(skt);
		recover_error("can't fdopen on '%s': %s", name, strerror(errno));
	}
	return fd;
}

flow_pos_t skf_getpos(flow_handle_t fh)
{
	return FLOW_ERROR;
}

int skf_setpos(flow_handle_t fh, flow_pos_t fp)
{
	return FLOW_ERROR;
}

int skf_rewind(flow_handle_t fh)
{
	return FLOW_ERROR;
}

char *skf_flowctl(flow_handle_t fh, char *ctl)
{
	file_flow_data *ffd = fh;
	int sock = fileno(ffd->x.hfile);
	if(!strcmp(ctl, "myaddr") || !strcmp(ctl, "peeraddr"))
	{
		dstring *tmp = ds_create(NULL);
		struct sockaddr_in sa;
		socklen_t salen = sizeof(sa);
		if((*ctl == 'm' ? getsockname : getpeername)(sock, (struct sockaddr *)&sa, &salen))
			recover_error("socket error: %s", strerror(errno));
		ds_appendstr(tmp, inet_ntoa(sa.sin_addr));
		ds_appendch(tmp, ':');
		ds_append(tmp, ds_fromint(ntohs(sa.sin_port), 10, 0));
		return DS_BODY(tmp);
	}
	else return ff_flowctl(fh, ctl);
}

static flow_class socket_flow_class = { "socket", skf_fopen, ff_close, ff_getc, ff_ungetc,
										ff_putc, ff_puts, skf_getpos, skf_setpos, skf_rewind,
										skf_flowctl
									  };
static flow_handle_t inetd_fopen(char *name, char *options, int writeable)
{
	file_flow_data *fd = ALLOC(file_flow_data);
	fd->writeable = 1;
	fd->preload   = 0;
	fd->undone 	  = FLOW_EOF;
	fd->x.hfile = stdin;
	return fd;
}

int inetd_putc(flow_handle_t fh, int ch)
{
	file_flow_data *ffd = fh;
	
	return fputc(ch, stdout) == EOF ? FLOW_ERROR : 0;
}

int inetd_puts(flow_handle_t fh, char *str)
{
	file_flow_data *ffd = fh;
	if(!str)
	{
		errno = EINVAL;
		return FLOW_ERROR;
	}
	return fputs(str, stdout) == EOF ? FLOW_ERROR : 0; 
}

static flow_class inetd_flow_class = { "inetd", inetd_fopen, ff_close, ff_getc, ff_ungetc,
									   inetd_putc, inetd_puts, skf_getpos, skf_setpos, skf_rewind,
									   skf_flowctl
};


void sockets_LTX_init(module *this)
{
	this_mod = this;
	register_new_flow("tcp", &socket_flow_class);
	register_new_flow("inetd", &inetd_flow_class);
}

static void bi_hostinfo(chunk *current, interpreter *i, value *args)
{
	create_storage(i, &hostinfo_storage_class, NULL, DS_BODY(args->val));
}

static void bi_hostname(chunk *current, interpreter *i, value *args)
{
	char buffer[256];
	gethostname(buffer, sizeof(buffer));
	ds_appendstr(get_pstack()->state->buffer, buffer);
}

/*
static master_instruction_src sockets_src[] = {
										 {NULL}
										};
*/										
static master_instruction_src sockets_src_add[] = {
										{"hostname", 1, bi_hostname, 0, 0, 0},
										{"hostinfo", 1, bi_hostinfo, 1, 1, 1},
										{NULL}
										};

void sockets_LTX_attach(mod_attach *ma)
{
//	ma->data = xmalloc(sizeof(sockets_src) * sizeof(instruction_t) / sizeof(*sockets_src));
//	replace_builtins(ma->i, sockets_src, ma->data);
	append_builtins(ma->i, sockets_src_add);
}

const char sockets_LTX_version[] = VERSION;
const char sockets_LTX_date[] = __DATE__;

/// \trailer[sockets]
